
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.127'
version = '1.2.127'
full_version = '1.2.127'
git_revision = '66ac370081054401fdfc9b51989c20e3900a03e1'
release = True

if not release:
    version = full_version
    short_version += ".dev"
