__author__ = 'labx'

"""
=========
Loop Management
=========

Widgets for loop management

"""

# Category description for the widget registry

NAME = "Basic Loops"

DESCRIPTION = "Widgets for Loop Management."

BACKGROUND = "#FAAC58"

ICON = "icons/loop_management.png"

PRIORITY = 2
